(*
  tactics.v
  Coq version: 8.0pl1 (Jul 2004)

  Tactics file
  
  Nadeem Abdul Hamid, September 2005
*)

Require Import Eqdep.



(**************************** Type level option ******************************)
Section optT.
Set Implicit Arguments.
Unset Strict Implicit.
Variable A : Type.
Inductive optT : Type :=
  | someT : A -> optT
  | noneT : optT.

Inductive listT : Type :=
  | nilT : listT
  | consT : A -> listT -> listT.

Fixpoint InT (a : A) (l : listT) {struct l} : Prop :=
  match l with
  | nilT => False
  | consT b m => b = a \/ InT a m
  end.

End optT.

(*********************** Custom grammar **************************************)

Infix "**" := prodT (at level 80, right associativity).
Infix ",*" := pairT (at level 80, right associativity).

(*********************** Custom definitions **********************************)

Definition injT1 A P (S : sigT P) : A := match S with
                                         | existT x _ => x
                                         end.

Definition injT2 A P (S : sigT P) : P (injT1 (A:=A) (P:=P) S) :=
  sigT_rect (fun s : sigT P => P (injT1 (A:=A) (P:=P) s))
    (fun (x : A) (p : P x) => p) S.








(*********************** Custom tactics **************************************)
Ltac iassumption :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH; eauto
  end.
Ltac iassumptionWith v :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH with v; eauto
  end.
Ltac iassumptionWith2 v1 v2 :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH with v1 v2; eauto
  end.
Ltac iassumptionWith3 v1 v2 v3 :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH with v1 v2 v3; eauto
  end.

Ltac splitIn hyp := elim hyp; clear hyp; intros.

Ltac splitAllHyps :=
  repeat
   match goal with
   |  (* splits all conjunctions in hypotheses *)
   H:(_ /\ _) |- _ => inversion_clear H; auto
   end.

Ltac concsplit :=
  repeat match goal with | |- _ /\ _ => split; auto end; try constructor.

Ltac mcrewrite v :=
  match goal with
  | H:(v = _) |- _ => rewrite H; auto
  | H:(_ = v) |- _ => rewrite <- H; auto
  | H:(v = _) |- _ => rewrite H; auto
  | H:(_ = v) |- _ => rewrite <- H; auto
  end.

Ltac mcrewrite2 v1 v2 := mcrewrite v1; mcrewrite v2; auto.

Ltac mcrewriteRev v :=
  match goal with
  | H:(v = _) |- _ => rewrite <- H; auto
  | H:(_ = v) |- _ => rewrite H; auto
  | H:(v = _) |- _ => rewrite <- H; auto
  | H:(_ = v) |- _ => rewrite H; auto
  end.

Ltac mcrewriteIn v t :=
  match goal with
  | H:(v = _),H1:t |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _) |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _ _) |- _ => rewrite H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _) |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(v = _),H1:t |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _) |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _ _) |- _ => rewrite H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _) |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _ _) |- _ => rewrite <- H in H1; auto
  end.

Ltac mcrewriteIn2 v t1 t2 := mcrewriteIn v t1; mcrewriteIn v t2.

Ltac mcrewriteIn3 v t1 t2 t3 :=
  mcrewriteIn v t1; mcrewriteIn v t2; mcrewriteIn v t3.

Ltac mcrewriteIn4 v t1 t2 t3 t4 :=
  mcrewriteIn v t1; mcrewriteIn v t2; mcrewriteIn v t3; mcrewriteIn v t4.

Ltac mcrewriteIn5 v t1 t2 t3 t4 t5 :=
  mcrewriteIn v t1; mcrewriteIn v t2; mcrewriteIn v t3; mcrewriteIn v t4;
   mcrewriteIn v t5.

Ltac mcrewriteRevIn v t :=
  match goal with
  | H:(v = _),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite H in H1; auto
  | H:(v = _),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite H in H1; auto
  end.


Ltac mcgen t :=
  match goal with
  | H:t |- _ => generalize H; clear H; auto
  | H:(t _) |- _ => generalize H; clear H; auto
  | H:(t _ _) |- _ => generalize H; clear H; auto
  | H:(t _ _ _) |- _ => generalize H; clear H; auto
  | H:(t _ _ _ _) |- _ => generalize H; clear H; auto
  | H:(t _ _ _ _ _) |- _ => generalize H; clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => generalize H; clear H; auto
  end.


Ltac mcgen_noclear t :=
  match goal with
  | H:t |- _ => generalize H; auto
  | H:(t _) |- _ => generalize H; auto
  | H:(t _ _) |- _ => generalize H; auto
  | H:(t _ _ _) |- _ => generalize H; auto
  | H:(t _ _ _ _) |- _ => generalize H; auto
  | H:(t _ _ _ _ _) |- _ => generalize H; auto
  | H:(t _ _ _ _ _ _) |- _ => generalize H; auto
  end.


Ltac mcgenapp1 t a1 :=
  match goal with
    | H:t |- _ => generalize (H a1); auto
    | H:(forall _ : t, _) |- _ => generalize (H a1); auto
    | H:(t _) |- _ => generalize (H a1); auto
    | H:(t _ _) |- _ => generalize (H a1); auto
    | H:(t _ _ _) |- _ => generalize (H a1); auto
    | H:(t _ _ _ _) |- _ => generalize (H a1); auto
    | H:(t _ _ _ _ _) |- _ => generalize (H a1); auto
    | H:(t _ _ _ _ _ _) |- _ => generalize (H a1); auto
  end.


Ltac mcgenapp2 t a1 a2 :=
  match goal with
  | H:t |- _ => generalize (H a1 a2); auto
  | H:(forall _ : t, _) |- _ => generalize (H a1 a2); auto
  | H:(forall _ : t, forall _, _) |- _ => generalize (H a1 a2); auto
  | H:(t _) |- _ => generalize (H a1 a2); auto
  | H:(t _ _) |- _ => generalize (H a1 a2); auto
  | H:(t _ _ _) |- _ => generalize (H a1 a2); auto
  | H:(t _ _ _ _) |- _ => generalize (H a1 a2); auto
  | H:(t _ _ _ _ _) |- _ => generalize (H a1 a2); auto
  | H:(t _ _ _ _ _ _) |- _ => generalize (H a1 a2); auto
  end.


Ltac mcgenapp3 t a1 a2 a3 :=
  match goal with
  | H:t |- _ => generalize (H a1 a2 a3); auto
  | H:(forall _ : t, _) |- _ => generalize (H a1 a2 a3); auto
  | H:(forall _ : t, forall _, _) |- _ => generalize (H a1 a2 a3); auto
  | H:(forall _ : t, forall _, forall _, _) |- _ => generalize (H a1 a2 a3); auto
  | H:(forall _ : t, forall _, forall _, forall _, _) |- _ => generalize (H a1 a2 a3); auto
  | H:(t _) |- _ => generalize (H a1 a2 a3); auto
  | H:(t _ _) |- _ => generalize (H a1 a2 a3); auto
  | H:(t _ _ _) |- _ => generalize (H a1 a2 a3); auto
  | H:(t _ _ _ _) |- _ => generalize (H a1 a2 a3); auto
  | H:(t _ _ _ _ _) |- _ => generalize (H a1 a2 a3); auto
  | H:(t _ _ _ _ _ _) |- _ => generalize (H a1 a2 a3); auto
  end.


Ltac mcgenapp4 t a1 a2 a3 a4 :=
  match goal with
  | H:t |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(forall _ : t, _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(forall _ : t, forall _, _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(forall _ : t, forall _, forall _, _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(forall _ : t, forall _, forall _, forall _, _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(t _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(t _ _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(t _ _ _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(t _ _ _ _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(t _ _ _ _ _) |- _ => generalize (H a1 a2 a3 a4); auto
  | H:(t _ _ _ _ _ _) |- _ => generalize (H a1 a2 a3 a4); auto
  end.


Ltac mcgenapp1clear t a1 :=
  match goal with
    | H:t |- _ => generalize (H a1); clear H; auto
    | H:(forall _ : t, _) |- _ => generalize (H a1); clear H; auto
    | H:(t _) |- _ => generalize (H a1); clear H; auto
    | H:(t _ _) |- _ => generalize (H a1); clear H; auto
    | H:(t _ _ _) |- _ => generalize (H a1); clear H; auto
    | H:(t _ _ _ _) |- _ => generalize (H a1); clear H; auto
    | H:(t _ _ _ _ _) |- _ => generalize (H a1); clear H; auto
    | H:(t _ _ _ _ _ _) |- _ => generalize (H a1); clear H; auto
  end.


Ltac mcgenapp2clear t a1 a2 :=
  match goal with
  | H:t |- _ => generalize (H a1 a2); clear H; auto
  | H:(forall _ : t, _) |- _ => generalize (H a1 a2); clear H; auto
  | H:(t _) |- _ => generalize (H a1 a2); clear H; auto
  | H:(t _ _) |- _ => generalize (H a1 a2); clear H; auto
  | H:(t _ _ _) |- _ => generalize (H a1 a2); clear H; auto
  | H:(t _ _ _ _) |- _ => generalize (H a1 a2); clear H; auto
  | H:(t _ _ _ _ _) |- _ => generalize (H a1 a2); clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => generalize (H a1 a2); clear H; auto
  end.


Ltac mcgenapp3clear t a1 a2 a3 :=
  match goal with
  | H:t |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(forall _ : t, _) |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(t _) |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(t _ _) |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(t _ _ _) |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(t _ _ _ _) |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(t _ _ _ _ _) |- _ => generalize (H a1 a2 a3); clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => generalize (H a1 a2 a3); clear H; auto
  end.



Ltac mcinv t :=
  match goal with
  | H:t |- _ => inversion H; auto
  | H:(t _) |- _ => inversion H; auto
  | H:(t _ _) |- _ => inversion H; auto
  | H:(t _ _ _) |- _ => inversion H; auto
  | H:(t _ _ _ _) |- _ => inversion H; auto
  | H:(t _ _ _ _ _) |- _ => inversion H; auto
  | H:(t _ _ _ _ _ _) |- _ => inversion H; auto
  end.


Ltac mcinvClear t :=
  match goal with
  | H:t |- _ => inversion_clear H; auto
  | H:(t _) |- _ => inversion_clear H; auto
  | H:(t _ _) |- _ => inversion_clear H; auto
  | H:(t _ _ _) |- _ => inversion_clear H; auto
  | H:(t _ _ _ _) |- _ => inversion_clear H; auto
  | H:(t _ _ _ _ _) |- _ => inversion_clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => inversion_clear H; auto
  end.

Ltac mcelim t :=
  match goal with
  | H:t |- _ => elim H; auto
  | H:(t _) |- _ => elim H; auto
  | H:(t _ _) |- _ => elim H; auto
  | H:(t _ _ _) |- _ => elim H; auto
  | H:(t _ _ _ _) |- _ => elim H; auto
  | H:(t _ _ _ _ _) |- _ => elim H; auto
  | H:(t _ _ _ _ _ _) |- _ => elim H; auto
  end.

Ltac mcelimClear t :=
  match goal with
  | H:t |- _ => elim H; clear H; auto
  | H:(t _) |- _ => elim H; clear H; auto
  | H:(t _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _ _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => elim H; clear H; auto
  end.


Ltac mcnewInd t :=
  match goal with
  | H:t |- _ => induction H; auto
  | H:(t _) |- _ => induction H; auto
  | H:(t _ _) |- _ => induction H; auto
  | H:(t _ _ _) |- _ => induction H; auto
  | H:(t _ _ _ _) |- _ => induction H; auto
  | H:(t _ _ _ _ _) |- _ => induction H; auto
  | H:(t _ _ _ _ _ _) |- _ => induction H; auto
  end.


Ltac mcclear t :=
  match goal with
    | H2:t |- _ => clear H2
    | H2:(t = _) |- _ => clear t H2
    | H2:(_ = t) |- _ => clear t H2
    | H2:(t = _) |- _ => clear t H2
    | H2:(_ = t) |- _ => clear t H2
    | H2:(t _) |- _ => clear H2
    | H2:(t _ _) |- _ => clear H2
    | H2:(t _ _ _) |- _ => clear H2
    | H2:(t _ _ _ _) |- _ => clear H2
    | H2:(t _ _ _ _ _) |- _ => clear H2
    | H2:(t _ _ _ _ _ _) |- _ => clear H2
    | H2:(t _ _ _ _ _ _ _) |- _ => clear H2
  end.

Ltac mcclearAll :=
  repeat
   match goal with
   | H:(?X1 = ?X2) |- _ => try mcclear X1; try mcclear X2
   end.


Ltac mcinjection t :=
  try
   match goal with
   | H1:(t = _),H2:(_ = t) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(t = _),H2:(t = _) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(_ = t),H2:(_ = t) |- _ =>
       rewrite <- H1 in H2; injection H2; intros
   | H1:(t = _),H2:(_ = t) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(t = _),H2:(t = _) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(_ = t),H2:(_ = t) |- _ =>
       rewrite <- H1 in H2; injection H2; intros
   end;
   try
     match goal with
       | H:(Some t = _) |- _ => injection H; intros
       | H:(_ = Some t) |- _ => injection H; intros
     end; 
   try
     match goal with
       | H:(existS _ _ t) = (existS _ _ _) |- _ 
	 => generalize (inj_pair2 _ _ _ _ _ H); intros; clear H
       | H:(existS _ _ _) = (existS _ _ t) |- _ 
	 => generalize (inj_pair2 _ _ _ _ _ H); intros; clear H
     end;
   auto.

Ltac mcclear2 t1 t2 := try mcclear t1; try mcclear t2.
Ltac mcclear3 t1 t2 t3 := try mcclear t1; try mcclear t2; try mcclear t3.
Ltac mcclear4 t1 t2 t3 t4 :=
  try mcclear t1; try mcclear t2; try mcclear t3; try mcclear t4.
Ltac mcclear5 t1 t2 t3 t4 t5 :=
  try mcclear t1; try mcclear t2; try mcclear t3; try mcclear t4;
   try mcclear t5.
Ltac mcclear6 t1 t2 t3 t4 t5 t6 :=
  try mcclear t1; try mcclear t2; try mcclear t3; try mcclear t4;
   try mcclear t5; try mcclear t6.
Ltac mcclear7 t1 t2 t3 t4 t5 t6 t7 :=
  try mcclear t1; try mcclear t2; try mcclear t3; try mcclear t4;
   try mcclear t5; try mcclear t6; try mcclear t7. Ltac unfoldSplit trm hyp := unfold trm in hyp; splitAllHyps.
Ltac introsSplitC := intros; splitAllHyps; try split.
Ltac introsSplitH := intros; splitAllHyps.
Ltac unfoldFold trm := unfold trm in |- *; fold trm in |- *. 
Ltac unfoldFoldIn trm hyp := unfold trm in hyp; fold trm in hyp; splitAllHyps.

Ltac mcunfoldIn v t :=
  match goal with
  | H:t |- _ => unfold v in H; auto
  | H:(t _) |- _ => unfold v in H; auto
  | H:(t _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _ _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _ _ _ _) |- _ => unfold v in H; auto
  end.
Ltac mcunfoldInSelf t := mcunfoldIn t t.
Ltac mcunfoldFoldIn v t :=
  match goal with
  | H:t |- _ => unfoldFoldIn v H; auto
  end.
Ltac mcunfoldIn2 v1 v2 t :=
  match goal with
  | H:t |- _ => unfold v1, v2 in H; auto
  end.
Ltac mcunfoldIn3 v1 v2 v3 t :=
  match goal with
  | H:t |- _ => unfold v1, v2, v3 in H; auto
  end.



Ltac elimex hyp var := elim hyp; clear hyp; intros var hyp; splitAllHyps.

Ltac elimex2 hyp var1 var2 := elimex hyp var1; elimex hyp var2.
Ltac elimex3 hyp v1 v2 v3 := elimex hyp v1; elimex hyp v2; elimex hyp v3.
Ltac elimex4 hyp v1 v2 v3 v4 :=
  elimex hyp v1; elimex hyp v2; elimex hyp v3; elimex hyp v4.
Ltac elimex5 hyp v1 v2 v3 v4 v5 :=
  elimex hyp v1; elimex hyp v2; elimex hyp v3; elimex hyp v4; elimex hyp v5.
Ltac elimex6 hyp v1 v2 v3 v4 v5 v6 :=
  elimex hyp v1; elimex hyp v2; elimex hyp v3; elimex hyp v4; elimex hyp v5;
   elimex hyp v6.
Ltac elimex7 hyp v1 v2 v3 v4 v5 v6 v7 :=
  elimex hyp v1; elimex hyp v2; elimex hyp v3; elimex hyp v4; elimex hyp v5;
   elimex hyp v6; elimex hyp v7.

Ltac mcelimex var t :=
  match goal with
  | H:t |- _ => elimex H var
  | H:(t _) |- _ => elimex H var
  | H:(t _ _) |- _ => elimex H var
  | H:(t _ _ _) |- _ => elimex H var
  | H:(t _ _ _ _) |- _ => elimex H var
  | H:(t _ _ _ _ _) |- _ => elimex H var
  | H:(ex _) |- _ => elimex H var
  | H:(ex _) |- _ => elimex H var
  end.
Ltac mcelimex2 var1 var2 t :=
  match goal with
  | H:t |- _ => elimex2 H var1 var2
  | H:(ex _) |- _ => elimex2 H var1 var2
  | H:(ex _) |- _ => elimex2 H var1 var2
  end.
Ltac mcelimex3 var1 var2 var3 t :=
  match goal with
  | H:t |- _ => elimex3 H var1 var2 var3
  | H:(ex _) |- _ => elimex3 H var1 var2 var3
  | H:(ex _) |- _ => elimex3 H var1 var2 var3
  end.
Ltac mcelimex4 var1 var2 var3 var4 t :=
  match goal with
  | H:t |- _ => elimex4 H var1 var2 var3 var4
  | H:(ex _) |- _ => elimex4 H var1 var2 var3 var4
  | H:(ex _) |- _ => elimex4 H var1 var2 var3 var4
  end.

Ltac elimexIn hyp var := elimex hyp var; exists var.

Ltac existsii a b := exists a; exists b.
Ltac existsiii a b c := existsii a b; exists c.
Ltac existsiv a b c d := existsiii a b c; exists d.

Ltac absurdEqBase :=
  match goal with
  | _:(?X1 <> ?X1) |- _ => absurd (X1 = X1); auto
  | _:(?X1 <> ?X1) |- _ => absurd (X1 = X1); auto
  end.
Ltac absurdEq :=
  match goal with
  | H1:(?X1 = ?X2),H2:(?X1 <> ?X2) |- _ =>
      rewrite H1 in H2; absurdEqBase
  | H1:(?X1 = ?X2),H2:(?X2 <> ?X1) |- _ =>
      rewrite H1 in H2; absurdEqBase
  | H1:(?X1 = ?X2),H2:(?X1 <> ?X2) |- _ =>
      rewrite H1 in H2; absurdEqBase
  | H1:(?X1 = ?X2),H2:(?X2 <> ?X1) |- _ =>
      rewrite H1 in H2; absurdEqBase
  | _:(?X1 <> ?X1) |- _ => absurd (X1 = X1); auto
  | _:(?X1 <> ?X1) |- _ =>
      absurd (X1 = X1); auto
  | H1:(?X1 = Some _),H2:(?X1 = None) |- _ =>
      rewrite H1 in H2; inversion H2; auto
  | H1:(?X1 = someT _),H2:(?X1 = noneT _) |- _ =>
      rewrite H1 in H2; inversion H2; auto 
  | H:(Some _ = None) |- _ => inversion H; auto
  | H:(None = Some _) |- _ => inversion H; auto
  | H:(noneT _ = someT _) |- _ => inversion H; auto
  | H:(someT _ = noneT _) |- _ => inversion H; auto
  end.

Ltac absurdity :=
  try match goal with
      | H:False |- _ => absurd False; auto
      end; try absurdEq.


Ltac eapplyinf :=
  match goal with
    | H:(forall _, ?t) |- (?t) => eapply H; eauto
    | H:(forall _ _, ?t) |- (?t) => eapply H; eauto
    | H:(forall _ _ _, ?t) |- (?t) => eapply H; eauto
    | H:(forall _ _ _ _, ?t) |- (?t) => eapply H; eauto
    | H:(forall _ _ _ _ _, ?t) |- (?t) => eapply H; eauto

    | H:(forall _, ?t _) |- (?t _) => eapply H; eauto
    | H:(forall _ _, ?t _) |- (?t _) => eapply H; eauto
    | H:(forall _ _ _, ?t _) |- (?t _) => eapply H; eauto
    | H:(forall _ _ _ _, ?t _) |- (?t _) => eapply H; eauto
    | H:(forall _ _ _ _ _, ?t _) |- (?t _) => eapply H; eauto

    | H:(forall _, ?t _ _) |- (?t _ _) => eapply H; eauto
    | H:(forall _ _, ?t _ _) |- (?t _ _) => eapply H; eauto
    | H:(forall _ _ _, ?t _ _) |- (?t _ _) => eapply H; eauto
    | H:(forall _ _ _ _, ?t _ _) |- (?t _ _) => eapply H; eauto
    | H:(forall _ _ _ _ _, ?t _ _) |- (?t _ _) => eapply H; eauto

    | H:(forall _, ?t _ _ _) |- (?t _ _ _) => eapply H; eauto
    | H:(forall _ _, ?t _ _ _) |- (?t _ _ _) => eapply H; eauto
    | H:(forall _ _ _, ?t _ _ _) |- (?t _ _ _) => eapply H; eauto
    | H:(forall _ _ _ _, ?t _ _ _) |- (?t _ _ _) => eapply H; eauto
    | H:(forall _ _ _ _ _, ?t _ _ _) |- (?t _ _ _) => eapply H; eauto

    | H:(forall _, ?t _ _ _ _) |- (?t _ _ _ _) => eapply H; eauto
    | H:(forall _ _, ?t _ _ _ _) |- (?t _ _ _ _) => eapply H; eauto
    | H:(forall _ _ _, ?t _ _ _ _) |- (?t _ _ _ _) => eapply H; eauto
    | H:(forall _ _ _ _, ?t _ _ _ _) |- (?t _ _ _ _) => eapply H; eauto
    | H:(forall _ _ _ _ _, ?t _ _ _ _) |- (?t _ _ _ _) => eapply H; eauto

    | H:(forall _, ?t _ _ _ _ _) |- (?t _ _ _ _ _) => eapply H; eauto
    | H:(forall _ _, ?t _ _ _ _ _) |- (?t _ _ _ _ _) => eapply H; eauto
    | H:(forall _ _ _, ?t _ _ _ _ _) |- (?t _ _ _ _ _) => eapply H; eauto
    | H:(forall _ _ _ _, ?t _ _ _ _ _) |- (?t _ _ _ _ _) => eapply H; eauto
    | H:(forall _ _ _ _ _, ?t _ _ _ _ _) |- (?t _ _ _ _ _) => eapply H; eauto
  end.

Ltac existSelim :=
  match goal with
    | H:(existS ?P ?x ?Px=existS ?P ?x ?Px') |- _ =>
      generalize (@Eqdep.inj_pair2 _ P x Px Px' H); clear H; intro; auto
  end.


Lemma inj_pairP2 :
  forall (U:Set) (P:U -> Prop) (p:U) (x y:P p),
    exist P p x = exist P p y -> x = y.
Proof.
  intros.
  apply (eq_dep_eq U P). 
  change p at 1 with (proj1_sig (exist P p x)).
  change x at 2 with (proj2_sig (exist P p x)).
  rewrite H.
  apply eq_dep_intro.
Qed.


Ltac existPelim :=
  match goal with
    | H:(exist ?P ?x ?Px=exist ?P ?x ?Px') |- _ =>
      generalize (@inj_pairP2 _ P x Px Px' H); clear H; intro; auto
  end.

